/*
 * =====================================================================================
 *
 *       Filename:  Depolarizer.h
 *
 *    Description:  Depolarizer abstract class and camac implementation
 *
 *        Version:  1.0
 *        Created:  03.10.2017 16:52:47
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */
#ifndef IBN_DEPOLARIZER_H
#define IBN_DEPOLARIZER_H

#include <chrono>
#include <string>
#include <forward_list>
#include <random>
#include <list>
#include <optional>

#include <thread>

#include "device/Depolarizer.h"
#include "device/Generator.h"
#include "device/DelayLine.h"
#include "device/Attenuator.h"

#include "barrier.h"

#include <iostream>
#include <type_traits>
#include <typeinfo>
#include <limits>




//constexpr inline auto get_frequency(double f, int n = 0, double f0 = 818924.485 ) -> double {
//	constexpr double RDM = 440.6484586602595;
//	return f == 0 ? 0 :(n + f/f0)*RDM;
//}

enum VerboseLevel {
	OFF   = 0,
	ERROR = 1,
	WARNING = 2,
	INFO = 3,
	ALL = 4,
};

class Depolarizer : public device::Depolarizer::State, public device::Depolarizer
{
	device::Info info;
	const double STEP_NUMBER_IN_WIDTH=5; //how many steps on depolarizer  fixed or dynamic width
	mutable double MAX_SCAN_TIME = 3600*5; //максимальное время сканирования
	public:
	const double MIN_FREQUENCY=5e4*Hz;
	double  initial_frequency = 500000;     //initial scan frequency, Hz
	//конечная частота сканирования, если не равна нулю, то
	//при её пересечении деполяризатором сканирование останавливается
	double  final_frequency = 0; //Hz            
	double revolution_frequency=818924.485*Hz; //revolution frequency
	VerboseLevel  verbose_level = VerboseLevel::ERROR;

	//typedef std::normal_distribution<>  pdf_t;
	typedef std::uniform_real_distribution<> pdf_t;
	std::unique_ptr<pdf_t> pdf; //frequency probability distribution function

	std::chrono::steady_clock::time_point begin_time;

	Depolarizer(std::string name, std::shared_ptr<device::Generator>, std::shared_ptr<device::Attenuator>, std::shared_ptr<device::DelayLine>);

	const device::Info & GetInfo(void) const final { return info; }

	~Depolarizer(void);

	virtual void    start_scan(void) final;
	virtual void     stop_scan(void) final ;
	virtual void    pause_scan(void) final ;
	virtual void continue_scan(void) final ;
	virtual bool       is_scan(void) const final { return isscan; };
	virtual bool         is_on(void) const final;
	virtual void           off(void) final; //turn off the generator
	virtual void            on(void) final; //turn on the generator
	virtual void            on(double f) final; //turn on the generator with spec frequency f in Hz


	virtual double get_revolution(void) const final { return revolution_frequency; }
	virtual int get_harmonic_number(void) const final { return harmonic_number; }
	virtual double get_frequency(void) const     final  {return frequency; };
	virtual double get_initial(void) const final { return initial_frequency; }; 
	virtual double get_final(void) const final { return final_frequency; };
	virtual double get_speed(void) const final { return speed; }
	virtual double get_width(void) const final { return width; }
	virtual double get_step(void) const final  { return step; }
	std::chrono::nanoseconds get_step_time(void) const { return step_time; }
	double get_delay(void) const;


	virtual void set_attenuation(double a) final; //set attenuation
	virtual void set_attenuation(void) final ; //set default attenuation
	virtual void set_harmonic_number(int n)  final  { harmonic_number = n; }
	virtual void set_revolution(double f) final { revolution_frequency = f; }



	virtual double get_attenuation(void) const final;

	void set_speed(double) final; 
	void set_step(double) final;
	void set_initial(double f) final;
	void set_final(double f) final;
	void set_frequency(double f) final;
	void set_width(double width) final;
	void set_delay(double delay_ns);



	void set_step_time(void);

	//double get_energy(double f /* frequeyncy Hz */ ) //return energy in MeV units
	//{
	//	return f == 0 ? 0 :(harmonic_number + f/revolution_frequency)*RDM;
	//}



	//************* frequency record part ****************
	public:
	bool is_record_freq = true;

	//typedef std::forward_list< DepolarizerState >  DepolarizerStateList; 
	typedef std::list< device::Depolarizer::State >  DepolarizerStateList; 
	DepolarizerStateList flist; //здесь хранится каждое состояние деполяризатора
	std::list<device::Depolarizer::State> state_change_list; //Журнал изменения состояния деполяризатора
	std::list< std::pair<device::Depolarizer::State, device::Depolarizer::State> > scan_list; //список сканирований
								  
	const DepolarizerStateList & get_frequency_list(void) const { return flist; } 
	void clear_frequency_list(void);

	device::Depolarizer::State get_state(void) const;

	std::optional<device::Depolarizer::State> get_state(const std::chrono::system_clock::time_point & ) const;

	mybarrier::count_variable  register_flist_reader(void) { return clear_barrier.registrate(); }
	void check_and_wait_flist_clear(void) { clear_barrier.wait(); }


	void set_verbose_level(VerboseLevel level) {
		verbose_level = level;
	}

	VerboseLevel get_verbose_level(void) const { return verbose_level; }

	void add_state_to_change_list(bool is_scan_flag);

	void set_max_scan_time(double t) const  { MAX_SCAN_TIME = t; }
	double get_max_scan_time(void) const  { return MAX_SCAN_TIME; }


	private:

	mybarrier clear_barrier;
	//DepolarizerStateList::iterator last_frequency_list_point_iterator;

	/* *******************************  end of frquency point list part ***************************************/


	void start_thread(void);
	void scan(void);


	std::atomic<bool> atomic_isscan;

	//volatile bool isscan; /*  flag for scan */
	std::thread  scan_thread; /* thread for scan*/
	std::chrono::nanoseconds step_time; //time for scan step, ns

	double local_ifreq;
	std::chrono::steady_clock::time_point local_begin_time;
	double last_frequency;

	std::shared_ptr<device::Generator>   generator; //synthesizer
	std::shared_ptr<device::Attenuator> attenuator; //attenuator device
	std::shared_ptr<device::DelayLine>  delay_line; //delay line
													//
	unsigned long scan_counter=0; //счетчик считает количество сканирований
};
#endif
