/*
 * =====================================================================================
 *
 *       Filename:  DepolarizerServer.cpp
 *
 *    Description:  Depolarizer DepolarizerServer implementation
 *
 *        Version:  1.0
 *        Created:  14.06.2017 16:58:05
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */
#include <iostream>
#include <fstream>

#include <stdexcept>
#include <thread>
#include <fmt/printf.h>
#include <fmt/format.h>
#include <fmt/ostream.h>
#include <fmt/chrono.h>

#include <boost/algorithm/string.hpp>

#include "Depolarizer.h"
#include "DepolarizerServer.h"

#include "proto/DepolarizerMessage.pb.h"
#include "proto/Message.h"

#include "print.h"

#include <chrono>
using namespace depolarizer_message_namespace;

enum State {
	ONOFF = 0x1,
	SCAN  = 0x2
};


DepolarizerServer::DepolarizerServer(int port) : basic_server(port)
{
	send_period = std::chrono::seconds(1);
}

DepolarizerServer::DepolarizerServer(int port, std::shared_ptr<device::Depolarizer> d) : DepolarizerServer(port)
{
	set_depolarizer(d);
}


void DepolarizerServer::set_depolarizer(std::shared_ptr<device::Depolarizer> d)
{
	fDepol = d;
	setup_actions();
}

void DepolarizerServer::setup_actions()
{
	/* Создадим таблицу для данного типа данных опредилим, как этот тип
	 * данных передавать деполяризатору и как брать из деполяризатора.
	 * Это позволит сократить switch код при обработке комманд серверу */
	set_get_actions[DepolarizerMessage::HARMONIC_NUMBER] = {
		[&](double data) { fDepol->set_harmonic_number(static_cast<int>(data));    },
		[&]()            { return static_cast<int>(fDepol->get_harmonic_number()); }
	};
	set_get_actions[DepolarizerMessage::REVOLUTION_FREQUENCY] = {
		[&](double data) { fDepol->set_revolution(data);    },
		[&]()            { return fDepol->get_revolution(); } 
	};
	set_get_actions[DepolarizerMessage::ATTENUATION] = {
		[&](double data) { fDepol->set_attenuation(data);    },
		[&]()            { return fDepol->get_attenuation(); } 
	};
	set_get_actions[DepolarizerMessage::FREQUENCY] = {
		[&](double data) { fDepol->set_frequency(data);    },
		[&]()            { return fDepol->get_frequency(); } 
	};
	set_get_actions[DepolarizerMessage::INITIAL] = {
		[&](double data) { fDepol->set_initial(data);    },
		[&]()            { return fDepol->get_initial(); } 
	};
	set_get_actions[DepolarizerMessage::FINAL] = {
		[&](double data) { fDepol->set_final(data);    },
		[&]()            { return fDepol->get_final(); } 
	};
	set_get_actions[DepolarizerMessage::SPEED] = {
		[&](double data) { fDepol->set_speed(data);    },
		[&]()            { return fDepol->get_speed(); } 
	};
	set_get_actions[DepolarizerMessage::STEP] = {
		[&](double data) { fDepol->set_step(data);    },
		[&]()            { return fDepol->get_step(); } 
	};
	set_get_actions[DepolarizerMessage::WIDTH] = {
		[&](double data) { fDepol->set_width(data);    },
		[&]()            { return fDepol->get_width(); } 
	};
	set_get_actions[DepolarizerMessage::DELAY] = {
		[&](double data) { std::dynamic_pointer_cast<Depolarizer>(fDepol)->set_delay(data);    },
		[&]()            { return std::dynamic_pointer_cast<Depolarizer>(fDepol)->get_delay(); } 
	};
	set_get_actions[DepolarizerMessage::LOG_LEVEL] = {
		[&](double data) { std::dynamic_pointer_cast<Depolarizer>(fDepol)->set_verbose_level(static_cast<VerboseLevel>(data));    },
		[&]()            { return static_cast<double>(std::dynamic_pointer_cast<Depolarizer>(fDepol)->get_verbose_level()); } 
	};
	set_get_actions[DepolarizerMessage::RDM_CONSTANT] = {
		[&](double) {},
		[&]()       { return std::dynamic_pointer_cast<Depolarizer>(fDepol)->RDM; } 
	};
	set_get_actions[DepolarizerMessage::MAX_SCAN_TIME] = {
		[&](double data) { std::dynamic_pointer_cast<Depolarizer>(fDepol)->set_max_scan_time(data);  },
		[&]()            { return static_cast<double>(std::dynamic_pointer_cast<Depolarizer>(fDepol)->get_max_scan_time()); } 
	};
}

static std::string log_head(void) {
	//return fmt::format("{:%Y-%m-%d %H:%M:%S}: DepolarizerServer: ", fmt::localtime(std::time(nullptr)));
	return "";
}

void fill_depolarizer_message_state(DepolarizerMessage::State * mst, const device::Depolarizer::State & st) 
{
	mst->set_timestamp(st.timestamp.time_since_epoch().count());
	mst->set_ison(st.ison);
	mst->set_isscan(st.isscan);
	mst->set_harmonic_number(st.harmonic_number);
	mst->set_frequency(st.frequency);
	mst->set_speed(st.speed);
	mst->set_step(st.step);
	mst->set_width(st.width);
	mst->set_attenuation(st.attenuation);
	//fmt::print("{:%F %T} {} {} {} {}\n", st.timestamp, st.harmonic_number, st.frequency, st.speed, st.attenuation); 
}

/* ======================= PROCEED CONNECTION  ======================== */
void DepolarizerServer::handle_connection(std::shared_ptr<ip::tcp::socket> socket)
{
	//socket->set_option(boost::asio::socket_base::timeout(boost::posix_time::seconds(30)));
	std::string hostname;
	try {
		hostname = socket->remote_endpoint().address().to_string();
		print_info("Открыто соединение", hostname);
		std::string s;
		unsigned  message_id=0;

		while(true) {
			auto msg = receive_message<DepolarizerMessage>(socket);
			DepolarizerMessage rmsg; //respond message
			rmsg.set_id(++message_id);//id текущего ответа
			rmsg.set_respond_id(msg.id()); //установим на какое сообщение отвечаем
			rmsg.set_respond_timestamp(msg.timestamp()); //и в какое время оно было отправлено
			rmsg.set_status(DepolarizerMessage::OK); //ответ по умолчанию хорош
			rmsg.set_command(msg.command());; //установим на какую команду овтечаем
			switch(msg.command()) {
				case  DepolarizerMessage::GET: 
					{
						rmsg.set_data_type(msg.data_type());
						switch ( msg.data_type() ) {
							case DepolarizerMessage::FMAP: 
								{
									bool is_send_fmap = true;
									send_message(socket, rmsg); //send OK and start send frequency
									send_states_list_to_client(socket,is_send_fmap);
									return;

								}
								break;
							case DepolarizerMessage::STATE: 
								{
									auto  rst = rmsg.add_states(); //подготовим новый элемент rst - respond state
									fill_depolarizer_message_state(rst, *std::dynamic_pointer_cast<Depolarizer>(fDepol));
									rst->set_timestamp( std::chrono::system_clock::now().time_since_epoch().count() );
								}
								break;
							case DepolarizerMessage::STATE_BY_TIME:
								{
									//мы получили массив в котором должны быть заполнены только времена (timestamps)
									//проходим по всем временам
									for (int i = 0; i <msg.states_size() ; ++i) {
										auto  rst = rmsg.add_states(); //подготовим новый элемент rst - respond state
										using namespace std::chrono;
										//извлекаем timestamp, преобразовав его в chrono::system_clock:time_point
										auto timestamp = msg.states(i).timestamp();
										//auto timestamp = s.timestamp();
										auto t = system_clock::time_point(system_clock::duration(timestamp));
										//далее по нему мы будем осуществлять поиск состояния деполяризацтора
										auto opt  = std::dynamic_pointer_cast<Depolarizer>(fDepol)->get_state(t);
										if ( opt ) {
											//fmt::print("Нашли время");
											fmt::print("Нашли частоту {:.0f} {} \n", opt.value().timestamp.time_since_epoch().count()*1e-9, opt.value().frequency);
											//нашли состояние, заполним сообщение
											fill_depolarizer_message_state(rst, opt.value());
										} else {
											fmt::print("Не нашли время {:.0f} \n", timestamp*1e-9);
											//не нашли состояние в списке состоянией, возращаем только timestamp
											//остальные поля остаются не инициализованными
											rst->set_timestamp(timestamp);
										}
									}
								}
								break;
							default: 
								{
									if( auto it = set_get_actions.find(msg.data_type()); it!= set_get_actions.end() ) {
										rmsg.set_data(it->second.get());
									} else {
										print_info("GET command", "Unknown data type", msg.data_type());
									}
								}
								break;
						}
					}
					break;

				case  DepolarizerMessage::SET: 
					{
						auto type = msg.data_type();
						rmsg.set_data_type(type);
						switch(type) {
							default: 
								{
									if( auto it = set_get_actions.find(msg.data_type()); it!= set_get_actions.end() ) {
										it->second.set( msg.data() );
										rmsg.set_data(it->second.get());
									} else {
										print_error("SET command", "Unknown data type",  msg.data_type());
									}
								}
								break;
						}
					}
					break;
				case DepolarizerMessage::START:
					if( ! fDepol->is_scan() ) {
						fDepol->start_scan();
						using namespace std::chrono;
						rmsg.set_data_type(DepolarizerMessage::STATUS);
						std::this_thread::sleep_for(10ms);
						print_info("START scan", std::dynamic_pointer_cast<Depolarizer>(fDepol)->get_state());
					} else {
						rmsg.set_status(DepolarizerMessage::BAD_COMMAND_RESULT);
						print_warning("Unable to START scan: already scanning now");
					}
					break;
				case DepolarizerMessage::STOP:
					if( fDepol->is_scan() ) {
						print_info("STOP scan", std::dynamic_pointer_cast<Depolarizer>(fDepol)->get_state());
						fDepol->stop_scan();
						rmsg.set_data_type(DepolarizerMessage::STATUS);
					} else {
						rmsg.set_status(DepolarizerMessage::BAD_COMMAND_RESULT);
						print_warning("Unable to STOP: no scanning now");
					}
					break;
				case DepolarizerMessage::PAUSE:
					if( fDepol->is_scan() ) {
						print_info("PAUSE scan");
						fDepol->pause_scan();
						rmsg.set_data_type(DepolarizerMessage::STATUS);
					} else {
						print_warning("Unable to PAUSE: no scanning now");
						rmsg.set_status(DepolarizerMessage::BAD_COMMAND_RESULT);
					}
					break;
				case DepolarizerMessage::CONTINUE:
					if( !fDepol->is_scan() ) {
						fDepol->continue_scan();
						rmsg.set_data_type(DepolarizerMessage::STATUS);
						using namespace std::chrono;
						std::this_thread::sleep_for(10ms);
						print_info("Continue scan", std::dynamic_pointer_cast<Depolarizer>(fDepol)->get_state());
					} else {
						rmsg.set_status(DepolarizerMessage::BAD_COMMAND_RESULT);
						print_warning("Unable to continue: already scanning");
					}
					break;
				case DepolarizerMessage::TURN_ON:
					if(fDepol) {
						fDepol->on();
						rmsg.set_data_type(DepolarizerMessage::STATUS);
						//std::cout << log_head() << "Turn ON depolarizer" << std::endl;
					}
					break;
				case DepolarizerMessage::TURN_OFF:
					if(fDepol) {
						fDepol->off();
						rmsg.set_data_type(DepolarizerMessage::STATUS);
						//std::cout << log_head() << "Turn OFF depolarizer" << std::endl;
					}
					break;
				default:
					rmsg.set_status(DepolarizerMessage_Error_UNKNOWN_COMMAND);
					//std::cerr << log_head() << "Unknown command\n";
					break;
			};
			send_message(socket,rmsg);
		}
	}
	catch( boost::system::system_error & e) {
		print_error(hostname, e.what());
	}
	catch(std::exception &e) {
		print_error(hostname, e.what());
	} 
	print_info("Закрыто соединение", hostname);
}


//void DepolarizerServer::start_send_frequency(std::shared_ptr<ip::tcp::socket>  sock, volatile bool & is_send)
//{
//	is_send = true;
//	std::thread([this,sock,&is_send](void) { send_frequency(sock, is_send); }).detach();
//}


void DepolarizerServer::send_states_list_to_client(std::shared_ptr<ip::tcp::socket>  sock, volatile bool & is_send) 
{
	std::string hostname;
	try {
		hostname = sock->remote_endpoint().address().to_string();
		typedef std::chrono::system_clock clock;
		auto fDepol = std::dynamic_pointer_cast<Depolarizer>(DepolarizerServer::fDepol);
		fDepol->register_flist_reader();
		const auto & M = fDepol->get_frequency_list();
		//Ожидаем когда что-то появится в списке
		while (M.empty()) {
			std::this_thread::sleep_for(send_period);
		};
		auto single_send = [&sock](const device::Depolarizer::State & depol_state) {
			DepolarizerMessage msg;
			msg.set_command(DepolarizerMessage::GET);
			msg.set_data_type(DepolarizerMessage::FMAP);
			msg.set_id(0);
			fill_depolarizer_message_state(msg.add_states(), depol_state);
			send_message(sock,msg);
		};
		//Как только появилось отправим клиенту и запишем в итератор
		//последние отправленные данные.
		//Тут надо заметить, что last всегда смотрит в валидные данные
		auto  last = M.begin(); 
		single_send(*last);
		while(true) {
			std::this_thread::sleep_for(send_period);
			//send current frequency if depolarizer is on and no scan
			if(!fDepol->is_scan() && fDepol->is_on()) {
				device::Depolarizer::State st;
				single_send(fDepol->get_state());
			}
			//отправляем новыую порцию данных. Надо посмотреть что есть за следующем
			//после last элементом. Если ничего нет, то ничего не отправляем, 
			//а если есть то отправляем. И при этом в last записываем последнюю
			//отправленную запись. То есть опять таки указатель на last всегда должен
			//быть валидным.
			for (auto it = std::next(last) ; it != M.end() ; ++it) {
				single_send(*it);
				last = it;
			}
			fDepol->check_and_wait_flist_clear();
		}
	}
	catch( boost::system::system_error & e) {
		boost::system::error_code ec = e.code();
		print_error("Отправка списка состояний деполяризатора (send_states_list_to_client)", hostname, e.what());
	}
	catch(std::exception & e) {
		print_error("Отправка списка состояний деполяризатора (send_states_list_to_client)", hostname, e.what());
	}
	print_info("Завершаю процедуру отправки списка состояний деполяризатора (send_states_list_to_client) для ", hostname);
}

template<typename ... Types>
void DepolarizerServer::print(std::ostream & os, int log_tag, const Types & ... msgs) const 
{
	if ( static_cast<int>(log_tag) <= static_cast<int>(std::dynamic_pointer_cast<Depolarizer>(fDepol)->get_verbose_level() )) {
		using namespace std::chrono;
		std::string fmt_string="{:%F %T}.{:3}: ";
		constexpr size_t N = sizeof...(msgs);
		if(N==0) return;
		for(size_t i=0; i<N-1; ++i) {
			fmt_string += "{}: ";
		}
		fmt_string += "{}";
		fmt_string += "\n";
		auto now = system_clock::now();
		auto round_now = round<milliseconds>(now);
		auto ms = (round_now.time_since_epoch() % 1s).count();
		fmt::print(os, fmt::runtime(fmt_string),now,ms , msgs...);
		os << std::flush;
	}
}

template<typename ... Types>
void DepolarizerServer::print_error(const Types & ... msgs)  const
{
	print(std::cerr, VerboseLevel::ERROR, "ERROR", msgs...);
}

template<typename ... Types>
void DepolarizerServer::print_warning(const Types & ... msgs) const
{
	print(std::clog, VerboseLevel::ERROR, "WARNING", msgs...);
}

template<typename ... Types>
void DepolarizerServer::print_info(const Types & ... msgs) const
{
	print(std::clog, VerboseLevel::ERROR, "INFO", msgs...);
}

