/*
 * =====================================================================================
 *
 *       Filename:  Crate.h
 *
 *    Description: Crate device for working with camac modules
 *
 *        Version:  1.0
 *        Created:  10.12.2019 13:56:26
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */
#ifndef IBN_CAMAC_CRATE_H
#define IBN_CAMAC_CRATE_H

#include <list>
#include <memory>
#include <algorithm>
#include <cstring>
#include <fmt/format.h>

#include "lsi6camac.h"

#include "../device/Device.h"
namespace camac 
{
	/* operate with camac crate device */
	class	Crate : public device::Device, public device::Info
	{
		public:
			Crate(int iface, int cr);
			const device::Info & GetInfo() const final { return info; };
			~Crate(void);
			class Error;
			int GetFd()	const {return fd;}
			int GetInterfaceNumber()	const {return interface;}
			int GetCrateNumber() 	const {return crate;}
			inline void C(void);
			inline void Z(void);
		private:
			const int interface; //interface number
			const int crate; //crate number
			int fd; //crate File Deskriptor
			device::Info info;
	};

	static std::list<std::shared_ptr<Crate>> CRATELIST;

	inline std::shared_ptr<Crate> get_crate(int interface_number, int crate_number, std::string ="") 
	{
		auto it =  std::find_if ( 
					CRATELIST.begin(), CRATELIST.end(),
					[interface_number, crate_number](const std::shared_ptr<Crate> & c) {
						return c->GetInterfaceNumber() == interface_number && c->GetCrateNumber() == crate_number ;
					} 
				);
		if(it != CRATELIST.end() ) return *it;
		auto crate = std::make_shared<Crate>(interface_number,crate_number);
		CRATELIST.push_back(crate);
		return crate;
	}


	class Crate::Error  :  public std::runtime_error
	{
		const int Errno;
		public:
		Error(std::string message, const device::Info & info):	
			std::runtime_error(fmt::format(R"({}: device type: "{}", model: "{}", address: "{}", name: "{}", error: "{}")", message, info.type, info.model, info.address, info.name, strerror(errno)) ),
			Errno(errno)
			{ }
		int GetErrno()	const { return Errno; }
	};


	inline Crate::Crate(const int iface, const int cr)	: 
		interface(iface),
		crate(cr)
	{
		info.type = "CAMAC/CRATE";
		info.model = "K0607/lsi6";
		info.address = "/dev/lsi6card" + std::to_string(interface+1) + "channel" + std::to_string(crate+1);
		fd = open(info.address.c_str(),O_RDWR);
		if( fd < 0 ) throw Crate::Error("Unable to initialize crate", info);
	}

	inline Crate::~Crate(void) { close(fd);}

	inline void Crate::Z(void)
	{
		int rc=CAM_Z(GetFd());
		if(rc==-1) throw Crate::Error("Unable to do Z operation",info);
	}

	inline void Crate::C(void)
	{
		int rc=CAM_C(GetFd());
		if(rc==-1) throw Crate::Error("Unable to do C operation",info);
	}
}
#endif
