#ifndef IBN_CAMAC_GENERATOR_H
#define IBN_CAMAC_GENERATOR_H
#pragma once

#include "Module.h"
#include "../device/Generator.h"

#include <thread>
#include <mutex>
#include <numbers>
#include <bitset>
#include <iostream>
#include <fmt/printf.h>
#include <math.h>

namespace camac 
{


	/*
	 * =====================================================================================
	 *        Class:  Generator_DDS500CAM
	 * 
	 *  Description: Synthesizer DDS500CAM. Разработчик блока Шубин Е.И. 
	 *  
	 *               
	 * 
	 *       Author:  Ivan Nikolaev
	 *      Created:  11.05.2004 15:53:46 NOVST
	 *     Revision:  none
	 * =====================================================================================
	 */
	class  Generator_DDS500CAM : protected Module, public device::Generator
	{
		static constexpr double INTERNAL_REFERENCE_FREQUENCY=30e6; //Hz
		static constexpr double EXTERNAL_REFERENCE_FREQUENCY=5e6; //Hz
		//static constexpr double DELAY_STEP=5; //ns
		static constexpr std::chrono::nanoseconds DELAY_STEP{5};
		public:

		int serial_number;
		int firmware_version;

		const device::Info & GetInfo(void) const final { return Module::GetInfo(); }

		enum LOG_LEVEL {
			OFF = 0,
			ERROR = 1,
			WARNING = 2,
			INFO = 3,
			DEBUG=4
		};
		template<typename ... Types>
			void print(std::ostream & os, int log_level, std::string fmt_str, const Types & ... msgs) const {
				if(log_level <= INFO) {
					std::string debug_str;
					switch (log_level) {
						case OFF:
							debug_str = "";
							break;
						case ERROR:
							debug_str = "ERROR";
							break;
						case WARNING:
							debug_str = "WARNING";
							break;
						case INFO:
							debug_str = "INFO";
							break;
						case DEBUG:
							debug_str = "DEBUG";
							break;
						default:
							debug_str = "";
							break;
					}
					using namespace std::chrono;
					auto now = system_clock::now();
					//fmt::print(os,"{:%F %T}.{:03}: {:7}: Generator_DDS500CAM: ", now, (round<milliseconds>(now).time_since_epoch() % 1s).count(), debug_str);
					fmt::print(os,fmt::runtime(fmt_str), msgs...);
				}
			}

		Generator_DDS500CAM(const camac::Address adr, std::string name):
			Module(adr,"Generator", "CAMAC/DDS500CAM",  name) {
				Reset(); //сброс из документации F9
				clear_status_register_bits(StatusRegister::Clear); //дополнительно очистим биты
				frequency=0;
				//SetExternalReferenceSignal(); //внешний опорный генератор
				SetInternalReferenceSignal();
				//print(std::clog, INFO, "Set external reference signal {} Hz\n", REF_FREQ);
				//print(std::clog, INFO, "Set internal reference signal {} Hz\n", REF_FREQ);
				set_delay_mode(); //включаем режим задержек
				//std::cout << "StatReg after set_delay_mode: " << StatReg << std::endl;
				using namespace std::chrono;
				SetGate(200ns); //Надо чтобы не зацепляло пучок в соседней сепаратрисе 1/4 оборота
				print(std::clog, INFO, "Set full gate: delay1+delay2 = {}\n", 200ns);
				auto data = AF<16>(15,1);
				serial_number = static_cast<int>( (data>>8).to_ulong() & 0xff);
				firmware_version = static_cast<int>( data.to_ulong() & 0xff);
				info.serial_number = fmt::format("{}", serial_number);
				info.version = fmt::format("{}",firmware_version);
				Off();
			}

		int GetSerialNumber() const { return serial_number; }
		int GetVersion() const { return firmware_version; }


		virtual ~Generator_DDS500CAM(void) { }

		inline void On(void) final;
		inline void Off(void) final;
		inline bool IsOn(void) const final {return ison;}
		inline void SetFrequency(double freq_hz) final;
		inline double GetFrequency(void) const;
		inline void GenerateFrequency(double f_hz, const std::chrono::duration<double, std::ratio<1,1>> & )  final; 

		enum class ReferenceSignal {
			Internal,
			External
		};

		inline void set_reference_signal( ReferenceSignal rs, unsigned int pll=1, double ref_freq_hz = EXTERNAL_REFERENCE_FREQUENCY);

		inline void SetInternalReferenceSignal(unsigned int pll=1); 
		inline void SetExternalReferenceSignal(unsigned int pll=1, double re_freq_hz = EXTERNAL_REFERENCE_FREQUENCY); 
		inline ReferenceSignal GetReferenceSignalType(void) const;
		inline double GetReferenceFrequency(void) const { return REF_FREQ; }

		inline void SetGate(std::chrono::nanoseconds gate_ns);
		inline void Reset(void); /*  reset all configurations */

		protected:

		enum  class Channel  
		{
			C0 = 0x40,
			C1 = 0x80,
			C0_C1 = 0x40 | 0x80
		};


		public:
		inline void set_phase(double phase);
		protected:

		inline void set_mult_factor(unsigned pll);

		inline void set_delay_mode(void);
		inline void set_phase_auto_synchronization(void);
		inline void adjust_phase_shift(void);
		inline void set_phase_shift(double phase);
		inline void set_work_channel(Channel ch);
		inline void set_gates (std::chrono::nanoseconds gate1, std::chrono::nanoseconds  gate2);
		inline void read_gates (std::chrono::nanoseconds & gate1, std::chrono::nanoseconds  & gate2);
		/* Write data from camac register to microscheme registers */
		inline void update_dds(void);

		inline void set_frequency(double freq);

		struct  StatusRegister 
		{
			static constexpr std::bitset<11>    Clear{0};
			static constexpr std::bitset<11>    InOut1{1 <<  0};
			static constexpr std::bitset<11>    InOut2{1 <<  1};
			static constexpr std::bitset<11>       In1{1 <<  2};
			static constexpr std::bitset<11>       In2{1 <<  3};
			static constexpr std::bitset<11> DelayMode{1 <<  6};
			static constexpr std::bitset<11> GenIntExt{1 <<  7};
			static constexpr std::bitset<11>       SW0{1 <<  8};
			static constexpr std::bitset<11>       SW1{1 <<  9};
			static constexpr std::bitset<11>        P2{1 << 10};
			static constexpr std::bitset<11>        P3{1 << 11};
		};

		std::bitset<11> StatReg;
		std::bitset<11> StatRegSave; //состояние статусного регистра, используемое для включения и выключения



		struct Register
		{
			template<unsigned NumberOfBytes, uint32_t Adr> 
			struct RegisterInfo
			{
				static constexpr uint32_t address  = Adr;
				static constexpr uint32_t size = NumberOfBytes;
				static constexpr uint32_t max_size = 4;
				static constexpr uint32_t max_address = 0x18;
				static_assert(1 <= size, "AD9958 register size must be larger or equal 1 byte");
				static_assert(size <= max_size, "AD9958 register size must be less or equal 4 bytes");
				static_assert(address <= max_address, "AD9958 register address must be less or equal 0x18");
			};
			//control registers
			static constexpr RegisterInfo<1, 0x00>  CSR{};
			static constexpr RegisterInfo<3, 0x01>  FR1{};
			static constexpr RegisterInfo<2, 0x02>  FR2{};
			//channel registers      Info  ,           
			static constexpr RegisterInfo<3, 0x03>  CFR{};
			static constexpr RegisterInfo<4, 0x04> CTW0{}; 
			static constexpr RegisterInfo<2, 0x05> CPW0{}; 
			static constexpr RegisterInfo<3, 0x06>  ACR{}; 
			static constexpr RegisterInfo<2, 0x07>  LSR{}; 
			static constexpr RegisterInfo<4, 0x08>  RDW{}; 
			static constexpr RegisterInfo<4, 0x09>  FDW{}; 
			//profile registers      Info         
			static constexpr RegisterInfo<4, 0x0A>  CTW1{};
			static constexpr RegisterInfo<4, 0x0B>  CTW2{}; 
			static constexpr RegisterInfo<4, 0x0C>  CTW3{}; 
			static constexpr RegisterInfo<4, 0x0D>  CTW4{}; 
			static constexpr RegisterInfo<4, 0x0E>  CTW5{}; 
			static constexpr RegisterInfo<4, 0x0F>  CTW6{}; 
			static constexpr RegisterInfo<4, 0x10>  CTW7{}; 
			static constexpr RegisterInfo<4, 0x11>  CTW8{};
			static constexpr RegisterInfo<4, 0x12>  CTW9{}; 
			static constexpr RegisterInfo<4, 0x13> CTW10{}; 
			static constexpr RegisterInfo<4, 0x14> CTW11{}; 
			static constexpr RegisterInfo<4, 0x15> CTW12{}; 
			static constexpr RegisterInfo<4, 0x16> CTW13{}; 
			static constexpr RegisterInfo<4, 0x17> CTW14{}; 
			static constexpr RegisterInfo<4, 0x18> CTW15{}; 
		};

		template < class Reg> 
		void write_register(Reg reg, std::bitset<8*Reg::size> data) 
		{
			//write register address to buffer register
			//AF(2,16, std::bitset<16>(reg.address));
			AF(2,16, std::bitset<16>(reg.address));

			//write data to buffer register
			
			const uint32_t d = static_cast<uint32_t>(data.to_ulong());

			AF(0,16, std::bitset<16>( d & 0xffff) );


			if constexpr ( Reg::size > 2) {
				AF(1,16, std::bitset<16>((d >> 16) & 0xffff) );
			}
			//write data from buffer register to AD9958 work register
			update_dds();
		}


		inline void write_status_register(std::bitset<11> );
		inline std::bitset<11> read_status_register(void) const;
		inline void set_status_register_bits(std::bitset<11>);
		inline void clear_status_register_bits(std::bitset<11>);


		double REF_FREQ;   //base frequency (internal or external). Опорная частота генератора.
		unsigned PLL; // multiplication factor. Множитель опорной частоты
		double FREQ_STEP; //минимальный шаг по частоте
		double MAX_FREQ; //максимально возможная частота
		double FREQ_STEP_INV; // коэффициент для преобразования частоты в 32bit код


		private:
		double frequency; /* generating frequency, Hz */
		double delay;  /*  delay in ns */
		double Time; /*  generation time in s. */
		bool  ison; /*  state of depolarizer */
		std::mutex mutex;
	};

	void Generator_DDS500CAM::On(void)
	{
		std::lock_guard lock(mutex);
		ison=true;
		set_status_register_bits(StatRegSave);//restore previose status register state
		//set_delay_mode(); //AF(5,16,0xc0);
		//AF(5,16,0xc0);
		set_frequency(frequency);
	}

	void Generator_DDS500CAM::Off(void)
	{
		std::lock_guard lock(mutex);
		ison=false;
		StatRegSave = read_status_register(); //save status register
		std::chrono::nanoseconds gate1, gate2;
		using namespace std::chrono;
		std::this_thread::sleep_for(100ms);
		read_gates(gate1, gate2);
		print(std::clog, DEBUG, "gate1 = {}, gate2 = {}\n", gate1, gate2);   
		clear_status_register_bits(StatusRegister::Clear); //AF(5,16,0x0);
		set_frequency(0);
	}

	void Generator_DDS500CAM::GenerateFrequency(double f /* Hz */, const std::chrono::duration<double, std::ratio<1,1>> & time) 
	{
		SetFrequency(f);
		std::this_thread::sleep_for(time);
	}

	void Generator_DDS500CAM::set_reference_signal( ReferenceSignal rs, unsigned int pll, double ref_freq_hz)
	{
		std::lock_guard lock(mutex);
		switch (rs ) {
			case ReferenceSignal::Internal:
				REF_FREQ=INTERNAL_REFERENCE_FREQUENCY;
				clear_status_register_bits(StatusRegister::GenIntExt);
				break;
			case ReferenceSignal::External:
				REF_FREQ = ref_freq_hz;
				set_status_register_bits(StatusRegister::GenIntExt);
				break;
			default:
				break;
		}
		set_mult_factor(pll);
	}

	void Generator_DDS500CAM::SetInternalReferenceSignal(unsigned int pll) 
	{
		set_reference_signal(ReferenceSignal::Internal,pll);
		print(std::clog,INFO, "SetInternalReferenceSignal: pll = {}, reference signal frequency = {:.12} MHz\n", PLL, REF_FREQ/1e6);
	}

	void Generator_DDS500CAM::SetExternalReferenceSignal(unsigned int pll, double ref_freq)
	{

		set_reference_signal(ReferenceSignal::External, pll, ref_freq);
		print(std::clog,INFO, "SetExternalReferenceSignal: pll = {}, reference signal frequency = {:.12} MHz\n", PLL, REF_FREQ/1e6);
	}

	Generator_DDS500CAM::ReferenceSignal Generator_DDS500CAM::GetReferenceSignalType(void) const {
		bool is_ext = (read_status_register() & StatusRegister::GenIntExt).any();
		if ( is_ext ) return ReferenceSignal::External;
		return ReferenceSignal::Internal;
	}


	void Generator_DDS500CAM::Reset(void) /*  reset all configurations */
	{
		std::lock_guard lock(mutex);
		StatReg=StatusRegister::Clear;
		AF(0,9);
	}

	void Generator_DDS500CAM::SetGate(std::chrono::nanoseconds gate)
	{
		std::lock_guard lock(mutex);
		delay=static_cast<double>(gate.count())/2.+20.0;
		set_gates(gate/2, gate/2);
		set_phase_auto_synchronization();
		adjust_phase_shift();
	}

	void Generator_DDS500CAM::set_frequency(double freq)
	{
		set_work_channel(Channel::C0_C1);//устанавливается частота одинаковая дли вдух каналов
		write_register(Register::CTW0, static_cast<uint32_t>(fabs(freq)*FREQ_STEP_INV));
		//uint32_t freq_kod=static_cast<uint32_t>(fabs(freq)*FREQ_STEP_INV); //  calculate code
		//uint32_t b10=freq_kod & 0xffff; /* low two byte */
		//uint32_t b32=freq_kod >> 16;    /* high tow byte */
		//AF(2,16,0x04); //write for frequency register;
		//AF(0,16,&b10);
		//AF(1,16,&b32);
		//update_dds();
		//set_phase_shift(-2.*M_PI*freq*delay*1e-9-M_PI);
		adjust_phase_shift();
	}


	/*
	 *--------------------------------------------------------------------------------------
	 *       Class:  Generagor_DDS500
	 *      Method:  set_frequency
	 * Description:  Independently writes frequency for two channel.
	 *--------------------------------------------------------------------------------------
	 */
	//void Generator_DDS500::set_frequency(double f1, double f2)
	//{
	//}

	void Generator_DDS500CAM::SetFrequency(double freq)
	{
		std::lock_guard lock(mutex);
		frequency = freq;
		if(ison) set_frequency(frequency);
	}

	double Generator_DDS500CAM::GetFrequency(void) const { return frequency; }


	void Generator_DDS500CAM::set_delay_mode(void) 
	{
		// Включить режим Задержек
		set_status_register_bits(StatusRegister::DelayMode);
	}


	void  Generator_DDS500CAM::set_phase_auto_synchronization(void)
	{
		//Взвести ит FR2<13>=1 для автоматической синхронизации каналов по фазе
		//AF(2,16,0x02); //write for address 0x02
		//AF(0,16,0x2000); // turn on bit FR2<13>=1
		//update_dds(); // UPDATE_DDS
		write_register(Register::FR2, 0x2000);
	}


	//записать фазовый сдвиг $\Delta \phi$ канала 1
	void Generator_DDS500CAM::set_phase_shift(double phase) 
	{
		// вычисление кода фазового сдвига:
		// kod = \Delta \phi / 2 \pi  * 2^14
		constexpr double KOEF = static_cast<double>(0x1<<14)/(2.*M_PI);
		set_work_channel(Channel::C1);
		write_register(Register::CPW0, static_cast<uint32_t>(phase*KOEF));
		//uint32_t kod = static_cast<uint32_t>(phase*KOEF);
		///* Control channel 1 */
		//AF(2,16,0x00); //write for address 0x00
		//AF(0,16,static_cast<uint32_t>(Channel::C1)); //turn on bit CSR<7>=1 channel 1
		//update_dds();
		//AF(2,16,0x05); //write for phase register
		//AF(0,16,&kod); //write code for phase
		//update_dds();
	}

	void Generator_DDS500CAM::set_phase(double phase) 
	{
		constexpr double KOEF = static_cast<double>(0x1<<14)/(2.*M_PI);
		set_work_channel(Channel::C0);
		write_register(Register::CPW0, static_cast<uint32_t>(phase*KOEF));
	}

	void Generator_DDS500CAM::adjust_phase_shift(void) 
	{
		set_phase_shift(2.*M_PI*frequency*delay*1e-9+M_PI);
	}


	void Generator_DDS500CAM::set_work_channel(Channel ch) 
	{
		write_register(Register::CSR, static_cast<uint8_t>(ch) ); 
		//AF(2,16,0x0); //write CSR for addres 0x00 //istruction byte
		//uint32_t data = static_cast<uint32_t>(ch);
		//AF(0,16, data);//turn on bit csr<6>=1 (channel0) csr<7>=1 (channel1)
		//update_dds();
	}

	void Generator_DDS500CAM::set_gates(std::chrono::nanoseconds gate1, std::chrono::nanoseconds  gate2)
	{
		using namespace std::chrono;
		if(gate1<0ns) gate1=0ns; 
		if(gate2<0ns) gate2=0ns; 
		std::bitset<16> code1(gate1/DELAY_STEP & 0xff);
		std::bitset<16> code2(gate2/DELAY_STEP & 0xff);
		//AF(3,16,static_cast<uint16_t>(code.to_ulong()));
		AF(3,16, code1 | (code2 << 8));
	}

	void Generator_DDS500CAM::read_gates(std::chrono::nanoseconds & gate1, std::chrono::nanoseconds  & gate2)
	{
		using namespace std::chrono;
		AF(3,26); //сдвиг данных из DS1023 (микросхемы задержек) в буферный сдвиговый регистр задержек DR
		std::bitset<16> data = AF<16>(3,0); //чтение буферного регистра 
		uint16_t data1 = static_cast<uint16_t>(0xff & data.to_ulong());
		uint16_t data2 = static_cast<uint16_t>(0xff & (data>>8).to_ulong());
		gate1 = 5ns*data1;
		gate2 = 5ns*data2;
	}

	/* Write data from camac register to microscheme registers */
	void Generator_DDS500CAM::update_dds(void) 
	{
		AF(0,26); //update dds
	}

	//void Generator_DDS500CAM::write_status_register(uint16_t data)
	void Generator_DDS500CAM::write_status_register(std::bitset<11> data)
	{
		AF(5,16,std::bitset<16>(data.to_ulong()));
	}

	std::bitset<11> Generator_DDS500CAM::read_status_register(void) const
	{
		//uint16_t data;
		//AF(5,0,&data);
		return std::bitset<11>(AF<16>(5,0).to_ulong());
	}

	void Generator_DDS500CAM::set_status_register_bits(std::bitset<11> mask) 
	{
		StatReg = read_status_register() | mask;
		write_status_register(StatReg);
	}

	void Generator_DDS500CAM::clear_status_register_bits(std::bitset<11> mask) 
	{
		StatReg = read_status_register() & mask;
		write_status_register(StatReg);
	}

	void Generator_DDS500CAM::set_mult_factor(unsigned k)
	{
		PLL = k;
		MAX_FREQ = PLL * REF_FREQ;
		FREQ_STEP = MAX_FREQ / ( uint64_t{1} << 32 );
		FREQ_STEP_INV = 1./FREQ_STEP;
	}
}

#endif
