/*
 * =====================================================================================
 *
 *       Filename:  Module.h
 *
 *    Description: Camac Module base interface
 *
 *        Version:  1.0
 *        Created:  12.12.2019 
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */
#pragma once
#ifndef IBN_CAMAC_MODULE_H
#define IBN_CAMAC_MODULE_H

#include "../device/Device.h"
#include "Crate.h"

#include <sstream> 
#include <iostream>
#include <fmt/format.h>
#include <bitset>

namespace camac
{
	//Description of camac module address
	struct Address
	{
		int interface=-1;  // camac interface in PC,
		int crate=-1;      // crate  in interface
		int position=-1;   // position  in crate
		int channel=-1;    // subaddress 
		Address(void) = default;
		Address(const Address &) = default;
		Address(std::string, char delim = ':');
		Address(int, int, int, int=-1);
		std::string str(void) const;
	};

	//Basic class for all camac blocks
	class Module :  public device::Device
	{
		protected:
			device::Info info;
		public:
			Module(const Address &, std::string type ="CAMAC/Module", std::string model ="unknown",std::string name ="unknown");

			Module(Crate & crate, int position, std::string type = "CAMAC/Module", std::string model = "unknown", std::string name = "unknown");

			virtual ~Module(void) { }

			virtual const device::Info & GetInfo(void) const { return info; }

			class Error;

			void CheckX(bool b=true) { check_X = b ; }

			int GetInterface()   const {return adr.interface; } //Get interface number
			int GetCrate() const { return adr.crate; } //Get crete number in interface
			int GetPosition()    const {return adr.position;} //Get position numb,er in crate
			int GetChannel()    const {return adr.channel;} //Get channel number in create
			inline void C();
			inline void Z();
			//LAM request
			inline bool IsLam(void) const ;
			inline double WaitLam(double sec) const ;

			inline bool IsLam(int, int ) const ;
			inline void WaitLam(int , int ) const;

			template<size_t Nbits>
			inline void AF(int A , int F, std::bitset<Nbits> data) 
			{
				static_assert( Nbits == 16 || Nbits == 24);
				int rc;
				if constexpr (Nbits == 16) {
					rc = CAMW(crate->GetFd(), NAF(adr.position,A,F), static_cast<unsigned int> (data.to_ulong()) );
				} else  if constexpr (Nbits == 24) {
					rc = CAM24W(crate->GetFd(), NAF(adr.position,A,F), static_cast<unsigned int>(data.to_ulong()) );
				} else {
					throw std::domain_error("AF bitset instruction allow only 16 or 24 bits");
				}
				SetXQ(rc);
				CheckError(A,F,"");
			}

			template<size_t Nbits>
			inline std::bitset<Nbits> AF(int A , int F) const 
			{
				static_assert( Nbits == 16 || Nbits == 24);
				std::bitset<Nbits> result;
				int rc;
				if constexpr (Nbits == 16) {
					unsigned int data;
					rc = CAM(crate->GetFd(), NAF(adr.position,A,F), &data );
					result = std::bitset<Nbits>(data);
				} else  if constexpr (Nbits == 24) {
					unsigned int data;
					rc = CAM24(crate->GetFd(), NAF(adr.position,A,F), &data );
					result = std::bitset<Nbits>(data);
				} else {
					throw std::domain_error("AF bitset instruction allow only 16 or 24 bits");
				}
				SetXQ(rc);
				CheckError(A,F,"");
				return result;
			}

			inline void AF(int A , int F) const {
				int rc;
				unsigned int data;
				rc = CAM(crate->GetFd(), NAF(adr.position,A,F), &data );
				SetXQ(rc);
				CheckError(A,F,"");
			}

			//inline void AF(int, int) const;
			//inline void AF(int, int, uint16_t *) const;
			//inline void AF(int, int, uint32_t *) const;
			//inline void AF(int, int, uint32_t );

			inline void BlockAF(int, int, uint16_t  *, int);
			inline void BlockAF(int, int, uint32_t  *, int);
			inline void SetAndLockI(void);
			inline void ClearAndLockI(void);
			inline void UnlockI(void);
			inline void SetI(void);
			inline void ClearI(void);
			inline bool GetI(void) const;
		protected:
			Address adr; //адресс
			std::shared_ptr<Crate> crate;  //Указатель на Crate
		private:	
			bool isinit;
			//Check errors
			std::string GetOperationInfo(const char * operation) const {
				return info.Type() + "/" +
					info.Model() + "/" +
					info.Name()  + "/" + 
					adr.str() + "/ " + 
					": RC=" + std::to_string(RC) + 
					", ERRNO=" + std::to_string(errno) +
					"(" + strerror(errno) + ")" +
					" operation: "+ operation +
					": X=" + std::to_string(X) +
					", Q=" + std::to_string(Q); 
			}
			void Check_Error(const std::string &func_name, int xq, int err) const;
			void CheckError(const char * operation) const;
			void CheckError(int A, int F, const char * operation) const;
			bool check_X=true; //check X answer
			bool XQ_ERROR_EXCEPTION = true;

			//status of last operation
			mutable bool X; 
			mutable bool Q;
			mutable int XQ;
			mutable int RC;

			void SetXQ(int rc)const  {
				RC = rc;
				XQ = 0x3 & (~rc);
				Q =  0x1 == (XQ & 0x1);
				X =  0x1 == ((XQ >> 1) & 0x1);
			}
	}; //end of class module

	/* implementation */

	inline Address::Address(int i, int c, int s, int ch) {
		interface = i;
		crate = c;
		position = s;
		channel = ch;
	}

	inline Address::Address(std::string s , char  )  { 
		char tmp;
		std::istringstream is(s);
		is >> interface >> tmp >> crate >> tmp>> position >> tmp >> channel;
	}

	inline std::string Address::str(void) const { return fmt::format("{}:{}:{}:{}",interface,crate,position, channel); }


	inline Module::Module(const Address &addr, std::string type, std::string model, std::string name)
	{
		adr = addr;
		info.type = type;
		info.model = model;
		info.name = name;
		info.address = adr.str();
		crate = get_crate(adr.interface, adr.crate);
		isinit = true;
	}


	inline void Module::CheckError(const char * operation="")  const {
		if(RC<0) {
			std::cerr << "ERROR: " << GetOperationInfo(operation) << ":  RC!=0 !!!" << std::endl;
			return;
		}
		if(!X && check_X) {
			std::cerr << "ERROR: " << GetOperationInfo(operation) << ":  X!=1 !!!" << std::endl;
		}
	}

	inline void Module::CheckError(int A, int F, const char * operation="")  const {
		char buf[1024];
		sprintf(buf,"A=%d, F=%d, %s",A,F,operation); 
		CheckError(buf);
	}

	// NAF instruction without data exchange
	//inline void Module::AF(int A, int F)  const {
	//	int rc = CAM(crate->GetFd(), NAF(adr.position,A,F), 0);
	//	SetXQ(rc);
	//	CheckError(A,F);
	//}

	//// NAF instruction with 16bit data
	//inline void Module::AF(int A, int F,uint16_t *DATA) const {
	//	uint32_t data = static_cast<uint32_t>(*DATA);
	//	int rc = CAM(crate->GetFd(), NAF(adr.position,A,F), &data);
	//	*DATA = static_cast<uint16_t>(data);
	//	SetXQ(rc);
	//	CheckError(A,F,"16bit");
	//}

	//// NAF instruction with 24bit data
	//inline void Module::AF(int A, int F, uint32_t *DATA) const {

	//	uint64_t data = static_cast<uint64_t>(*DATA);
	//	int rc = CAM24(crate->GetFd(), NAF(adr.position,A,F), &data);
	//	*DATA = (uint32_t)data;
	//	SetXQ(rc);
	//	CheckError(A,F,"24bit");
	//}

	//inline void Module::AF(int A, int F, uint32_t DATA) {
	//	AF(A,F,&DATA);
	//}

	inline void  Module::Z(void) {
		CAM_Z(crate->GetFd());
	}

	inline void Module::C(void) {
		CAM_C(crate->GetFd());
	}

	/*
	 *--------------------------------------------------------------------------------------
	 *       Class:  Module
	 *      Method:  WaitLam
	 * Description:  Wait the LAM request
	 * timeout > 0 - wait LAM during timeout seconds. If LAM happens it returns remain of timeout.  
	 *		     Zero value is possible. 
	 * timeout = 0 - check the LAM and return 0 if LAM happens or -1 otherwise.
	 * timeout < 0 - wait LAM forever.
	 * If no LAM during timeout ETIME error is returned.
	 * Waiting for timeout might be interrupt by unmasking signal. Corresponding error is returned.
	 *--------------------------------------------------------------------------------------
	 */
	inline double Module::WaitLam(double timeout)  const 
	{
		//ATTENTION architecture dependent part timeout*100 - is number of ticks 
		//for x86 architecture 1tick = 10ms
		static double tick_time=1e-2;
		int rc = CAM_LWAIT(crate->GetFd(), adr.position, int(timeout/tick_time));
		SetXQ(rc);
		CheckError("WaitLam");
		return double(rc)*1e-2;
	}


	/*
	 *--------------------------------------------------------------------------------------
	 *       Class:  Module
	 *      Method:  IsLam
	 * Description:  Check the LAM and return true if LAM or false otherwise
	 *--------------------------------------------------------------------------------------
	 */
	inline bool Module::IsLam(void)  const  {

		int rc = CAM_LWAIT(crate->GetFd(), adr.position, 0);
		if(rc==0) return true;
		return false;
	}



	inline bool Module::IsLam(int A, int F)  const {
		int rc = CAM(crate->GetFd(), NAF(adr.position,A,F), 0);
#ifdef PCI_CAMAC
		rc = 0x1 & ~rc;
#endif 
		return rc & 0x1;
	}

	inline void Module::WaitLam(int A, int F)  const {
		while(!IsLam(A,F)) usleep(10000);
	}

	//Установка бита I
	inline void Module::SetI(void) {
		CAM_ION(crate->GetFd());
	}

	//Сброс бита I
	inline void Module::ClearI() {
		CAM_IOFF(crate->GetFd());
	}

	//Получение бита I
	inline bool Module::GetI(void)	const {
		int rc = CAM_I(crate->GetFd());
		if(rc == 0 || rc ==1) return rc;
		Check_Error("GetI",rc,errno);
		return 0;
	}
} //end namespace camac


#endif //IBN_CAMAC_MODELU
