/*
 * =====================================================================================
 *
 *       Filename:  DepolarizerClient.h
 *
 *    Description:  
 *
 *        Version:  1.0
 *        Created:  23.12.2019 19:07:53
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */

#ifndef IBN_DEPOLARIZER_CLIENT_H
#define IBN_DEPOLARIZER_CLIENT_H

#include <string>
#include <chrono>
#include <functional>
#include <thread>
#include <boost/asio.hpp>
#include "../proto/DepolarizerMessage.pb.h"
#include "../proto/Message.h"

#include "../device/Depolarizer.h"


class DepolarizerClient  : public device::Depolarizer 
{
  mutable device::Depolarizer::State state;
  boost::asio::io_service ios;
  std::shared_ptr<boost::asio::ip::tcp::endpoint> address;
  mutable std::shared_ptr<boost::asio::ip::tcp::socket> socket;
  mutable boost::asio::streambuf streambuf;

  enum State {
    ONOFF = 0x1,
    SCAN  = 0x2
  };

  device::Info info;
  public:

  const device::Info & GetInfo(void) const  final {
	  return info;
  }

  DepolarizerClient(std::string host = "127.0.0.1", int port = 9090, std::string name = "")
  {
	info.type = "Depolarizer";
	info.model = "NetworkDepolarizer";
	info.address = host + ":" + std::to_string(port);
	info.name = name;
    connect(host,port);
  }

  void connect(std::string host, int port) {
    address = std::make_shared<boost::asio::ip::tcp::endpoint>(boost::asio::ip::address::from_string(host), port);
    socket = std::make_shared<boost::asio::ip::tcp::socket>(ios);
    socket -> connect(*address);
  }

  void Disconnect(void) { socket->close(); }


  virtual ~DepolarizerClient(void) { socket->close(); }



  void start_scan()    final { simple_command(DepolarizerMessage::START);   } 
  void stop_scan()     final { simple_command(DepolarizerMessage::STOP);    } 
  void pause_scan()    final { simple_command(DepolarizerMessage::PAUSE);   } 
  void continue_scan() final { simple_command(DepolarizerMessage::CONTINUE);} 
  void on()            final { simple_command(DepolarizerMessage::TURN_ON); } 
  void on(double f)    final 
  { 
	  simple_command(DepolarizerMessage::TURN_ON);
	  set_frequency(f);
  } //turn on the generator with spec frequency f in Hz
  void off()           final { simple_command(DepolarizerMessage::TURN_OFF);} 

  void update_state()  const {
    using namespace depolarizer_message_namespace;
    DepolarizerMessage msg;
    msg.set_id(message_id++);
    msg.set_command(DepolarizerMessage::GET);
    msg.set_data_type(DepolarizerMessage::STATE);
    send_message(socket, msg);
    msg = receive_message<DepolarizerMessage>(socket);
	if (msg.states_size()!=1) throw std::runtime_error("Wrong respond on GET STATE request: state_size != 1");
	auto & st = msg.states(0);
	state.timestamp = std::chrono::system_clock::time_point(std::chrono::system_clock::duration(st.timestamp()));
	state.ison = st.ison();
	state.isscan = st.isscan();
	state.harmonic_number = st.harmonic_number();
	state.frequency       = st.frequency();
	state.speed          = st.speed();
	state.step = st.step();
	state.width = st.width();
	state.attenuation = st.attenuation();
  }

  bool is_scan(void) const  {
	update_state();
	return state.isscan;
  };

  bool is_on() const {
	update_state();
	return state.ison;
  }



  void set_attenuation(double a)    final { set_value(DepolarizerMessage::ATTENUATION, a); }
  void set_speed(double a)          final { set_value(DepolarizerMessage::SPEED, a); }
  void set_step(double a)           final { set_value(DepolarizerMessage::STEP, a); }
  void set_width(double a)          final { set_value(DepolarizerMessage::WIDTH, a); }
  void set_delay(double a)          final { set_value(DepolarizerMessage::DELAY, a); }
  void set_initial(double a)        final { set_value(DepolarizerMessage::INITIAL, a); }
  void set_final(double a)          final { set_value(DepolarizerMessage::FINAL, a); }
  void set_frequency(double a)      final { set_value(DepolarizerMessage::FREQUENCY, a); }
  void set_revolution(double a)      final { set_value(DepolarizerMessage::REVOLUTION_FREQUENCY, a); }
  void set_harmonic_number(int n) final { set_value(DepolarizerMessage::HARMONIC_NUMBER, n); }

  double get_attenuation(void)     const final  { return get_value(DepolarizerMessage::ATTENUATION); }
  double get_speed(void)           const final  { return get_value(DepolarizerMessage::SPEED); }
  double get_step(void)            const final  { return get_value(DepolarizerMessage::STEP); }
  double get_width(void)            const final  { return get_value(DepolarizerMessage::WIDTH); }
  double get_delay(void)            const final  { return get_value(DepolarizerMessage::DELAY); }
  double get_initial(void)         const final  { return get_value(DepolarizerMessage::INITIAL); }
  double get_final(void)           const final  { return get_value(DepolarizerMessage::FINAL); }
  double get_frequency(void)       const final  { return get_value(DepolarizerMessage::FREQUENCY); }
  int get_harmonic_number(void) const final  { return static_cast<int>(get_value(DepolarizerMessage::HARMONIC_NUMBER)); }
  double get_revolution(void) const final {return get_value(DepolarizerMessage::REVOLUTION_FREQUENCY); }
 

  private:

  void simple_command(DepolarizerMessage_Command command) {
    using namespace depolarizer_message_namespace;
    DepolarizerMessage msg;
    msg.set_id(message_id++);
    msg.set_command(command);
    send_message(socket, msg);
  }

  void set_value(DepolarizerMessage_DataType type,  double a) {
    using namespace depolarizer_message_namespace;
    DepolarizerMessage msg;
    msg.set_id(message_id++);
    msg.set_command(DepolarizerMessage::SET);
    msg.set_data_type(type);
    msg.set_data(a);
    send_message(socket, msg);
  };

  double  get_value(DepolarizerMessage_DataType type) const {
    using namespace depolarizer_message_namespace;
    DepolarizerMessage msg;
    msg.set_id(message_id++);
    msg.set_command(DepolarizerMessage::GET);
    msg.set_data_type(type);
    send_message(socket, msg);
    msg = receive_message<DepolarizerMessage>(socket);
    return msg.data();
  };

  mutable uint32_t message_id = 0; 
};


#endif
