#ifndef IBN_DEVICE_ATTENUATOR_ARRAY_H
#define IBN_DEVICE_ATTENUATOR_ARRAY_H

#include "Attenuator.h"

#include <memory>
#include <list>
#include <map>
#include <limits>
#include <exception>

namespace device 
{
	class ArrayAttenuator  : public device::Attenuator
	{
		device::Info info;
		//use multimap to sort attenuators by it's attenuation
		typedef std::unique_ptr<Attenuator> att_ptr;
		typedef  std::multimap<double, att_ptr ,std::greater<double> > amap_t ;
		amap_t amap;
		double att;
		double att_quant;
		double MAX_ATTENUATION;
		public:
		virtual ~ArrayAttenuator(void)  { }
		ArrayAttenuator(std::string name="");

		void   SetAttenuation(double a) final;
		double GetAttenuation(void) const final; 
		double GetMaxAttenuation(void) const final { return MAX_ATTENUATION; }
		double GetStep(void)const  final { return att_quant; }
		const Info & GetInfo(void) const final { return info; }

		void AddAttenuator(std::unique_ptr<device::Attenuator> && ptr );
		void AddAttenuator(device::Attenuator * a);
		std::list < Attenuator* > GetAttenuators(void);
	};


	ArrayAttenuator::ArrayAttenuator(std::string name) 
	{
		info.type = "Attenuator";
		info.model = "ARRAY";
		info.address = "";
		info.name = name;
		MAX_ATTENUATION=0;
		att_quant=0;
	}

	void ArrayAttenuator::AddAttenuator(att_ptr && ptr )
	{
		if(!ptr) return;
		amap.emplace(std::make_pair(ptr->GetStep(),std::move(ptr)));
		//recalculate maximum attenuation and minimum attenuation step
		double max=0; 
		double quant = std::numeric_limits<double>::max();
		for(auto & item : amap ) {
			auto & device = item.second;
			quant=std::min(quant,device->GetStep());
			max+=device->GetMaxAttenuation();
		}
		MAX_ATTENUATION=max;
		att_quant=quant;
	}

	void ArrayAttenuator::AddAttenuator(device::Attenuator * a) {
		AddAttenuator(std::unique_ptr<device::Attenuator>(a));
	}

	/* Complex attenuation set.
	 * List of attenuators are sorted by decreasing of attenuation step.
	 * Desired attenuation is rounded to possible attenuation for first attenuation in queue.
	 * remains go to the next attenuation in queue
	 */
	void ArrayAttenuator::SetAttenuation(double a)   
	{
		if(amap.empty()) {
			att=0;
			return;
		}
		double s=0,m=0,sub=0,save;
		a= a >=0 ? a : -a ;
		save=a;
		if(a>MAX_ATTENUATION) a=MAX_ATTENUATION;
		for(auto & item : amap ) {
			s=item.first;
			m=item.second->GetMaxAttenuation();
			sub = unsigned(a/s)*s;
			if( sub > m) sub=m;
			item.second->SetAttenuation(sub);
			a-=sub;
		}
		att=save-a;
		if (a>att_quant) {
			throw std::runtime_error("Bad attenuation in ArrayAttenuator::SetAttenuation (a>att_quant)");
		}
	}

	double ArrayAttenuator::GetAttenuation(void) const 
	{
		size_t size=amap.size();
		if(size==0) return 0;
		double sum=0;
		for(auto & a : amap ) {
			sum+=a.second->GetAttenuation();
		}
		return sum;
	}

	std::list < device::Attenuator * > ArrayAttenuator::GetAttenuators(void) 
	{
		std::list <device::Attenuator* > result;
		for(auto & item :  amap ) {
			result.push_back(item.second.get());
		}
		return result;
	}
}
#endif
