import sys
import signal
from DepolarizerMessage_pb2 import *
import socket
import struct
import time, datetime
from threading import Thread

from math import modf


def frequency2energy(f,n) :
    # f is the frequency, Hz
    # n is harmonic number
    # R = 2*m_e/(g-2) 
    R = 440.64845866025958 #MeV  // PDG-2021
    F0 = 818924.126 #Hz = 181 801 156 Hz / 222 сепаратрисы
    if f == 0 : return 0.0
    return (f/F0+n)*R


def energy2frequency(E) :
    # E is the energy, MeV
    # R = 2*m_e/(g-2) 
    R = 440.64845866025958 #MeV  // PDG-2021
    F0 = 818924.126 #Hz = 181 801 156 Hz / 222 сепаратрисы
    n = E//R #harmonic number
    print("harmonic number = ", n)
    return F0*(E/R-E//R)

class Depolarizer:
    #вспомогательные функции для работы с сокетом
    # соединение
    def connect(host, port):
      sock = socket.socket()
      sock.connect((host, port))
      return sock

    def reconnect(self):
        self.sock.close()
        self.sock = Depolarizer.connect(self.host, self.port);

    def is_socket_open(sock):
        if sock is None:
            return False
        try:
            # Пробуем выполнить неблокирующую операцию
            sock.getpeername()
            return True
        except socket.error:
            return False

    #Упаковка (сериализация) и отравка сообщения DepolarizerMessage
    def send(sock, message):
        def send_msg(sock, msg):
            msg.timestamp = int(time.time()*1e9)
            sock.send(struct.pack('L',msg.ByteSize()))
            sock.send(msg.SerializeToString())
        try :
            send_msg(sock, message)
        except Exception as e: #лишь однократно отрабатываем обрыв соединения
            print("Ошибка в отправке запроса серверу деполяризатора",e)
            raise

    #Получение и десериализация сообщения DepolarizerMessage
    def receive(sock):
        def get_data(sock):
            #Получим данные с длиной сообщения
            data = sock.recv(8, socket.MSG_WAITALL)
            #расшифруем
            msg_len  = struct.unpack('L',data)[0]
            #получим само сообщение
            data = sock.recv(msg_len, socket.MSG_WAITALL)
            return data
        try :
            data = get_data(sock)
        except Exception as e: #лишь однократно отрабатываем обрыв соединения
            print("Ошибка получении ответа от сервера деполяризатора",e)
            raise
        m = DepolarizerMessage()
        try:
            m.ParseFromString(data)
        except Exception as e:
            print('Не могу распознать сообщение от деполяризатора')
            raise
        return m


    #Замечание: все операции с частотами (текущая, начальная, конечная, скорость, шаг, ширина)
    #выражены в Гц, а не MeV.
    #Таблица для генерации методов c обменом данными
    # set_<name>(data) get_<name>()
    data_types_mapping = {
            "frequency"            : DepolarizerMessage.FREQUENCY,
            "initial"              : DepolarizerMessage.INITIAL,
            "final"                : DepolarizerMessage.FINAL,
            "step"                 : DepolarizerMessage.STEP,
            "width"                : DepolarizerMessage.WIDTH,
            "speed"                : DepolarizerMessage.SPEED,
            "attenuation"          : DepolarizerMessage.ATTENUATION,
            "harmonic_number"      : DepolarizerMessage.HARMONIC_NUMBER,
            "revolution_frequency" : DepolarizerMessage.REVOLUTION_FREQUENCY,
            "log_level"            : DepolarizerMessage.LOG_LEVEL,
            "delay"                : DepolarizerMessage.DELAY,
            "max_scan_time"        : DepolarizerMessage.MAX_SCAN_TIME
            }

    #таблица для генерации методов без обмена данным
    do_types_mapping = {
            "start_scan"    : DepolarizerMessage.START,
            "stop_scan"     : DepolarizerMessage.STOP,
            "continue_scan" : DepolarizerMessage.CONTINUE
            }


    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.sock = Depolarizer.connect(self.host, self.port)
        self.message_id=0
        #Создаем getter и сеттеры для разных типов данных
        for name, n in self.data_types_mapping.items():
            setattr(self, f'get_{name}', self._create_getter(n))
            setattr(self, f'set_{name}', self._create_setter(n))

        for name, number in self.do_types_mapping.items():
            setattr(self, f'{name}', lambda n=number :  self.do(n) )

    #Вспомогательный метод для выполнения команды,
    #который не требует обмена данными
    #Возвращает статус операции 
    def do(self, command):
      m = DepolarizerMessage()
      m.id = self.message_id
      self.message_id = self.message_id + 1
      m.command = command
      Depolarizer.send(self.sock, m)
      m = Depolarizer.receive(self.sock)
      return m.status == DepolarizerMessage.OK

    #Вспомогательный метод
    #Получение числа определенного типа данных
    def get(self,data_type):
        if not Depolarizer.is_socket_open(self.sock): self.reconnect()
        m = DepolarizerMessage()
        m.id = self.message_id
        self.message_id = self.message_id + 1
        m.command = DepolarizerMessage.GET
        m.data_type = data_type
        Depolarizer.send(self.sock,m)
        m = Depolarizer.receive(self.sock)
        return m.data
    

    #Вспомогательный метод
    #Отправка числа определенного типа данных
    def set(self,data_type,data):
        if not Depolarizer.is_socket_open(self.sock): self.reconnect()
        #if not self.sock: self.reconnect()
        m = DepolarizerMessage()
        m.id = self.message_id
        self.message_id = self.message_id + 1
        m.command = DepolarizerMessage.SET
        m.data_type = data_type
        m.data = data;
        Depolarizer.send(self.sock, m)
        m = Depolarizer.receive(self.sock)
        return m.data

    def _create_getter(self, number):
        def getter():
            return self.get(number)
        return getter

    def _create_setter(self, number):
        def setter(data):
            return self.set(number, data)
        return setter
    
    def get_rdm_constant(self):
        return self.get(DepolarizerMessage.RDM_CONSTANT)

    def get_state(self):
        m = DepolarizerMessage()
        m.id = self.message_id
        self.message_id = self.message_id + 1
        m.command = DepolarizerMessage.GET
        m.data_type = DepolarizerMessage.STATE
        Depolarizer.send(self.sock, m)
        m = Depolarizer.receive(self.sock)
        return m.states[0]

    def is_off(self):
        return not self.get_state().ison

    def is_on(self):
        return self.get_state().ison

    def is_scan(self):
        return self.get_state().isscan

    def get_state_by_time(self, time_list ):
        m = DepolarizerMessage()
        m.id = self.message_id
        self.message_id = self.message_id + 1
        m.command = DepolarizerMessage.GET
        m.data_type = DepolarizerMessage.STATE_BY_TIME
        if type(time_list) is float:
            fp = m.states.add()
            fp.timestamp = int(1e9*time_list) #convert to nanoseconds
        else:
            for  t in time_list:
                fp = m.states.add()
                fp.timestamp = int(1e9*t) 
        Depolarizer.send(self.sock, m)
        m = Depolarizer.receive(self.sock)
        return m.states

    #def signal_handler(self, signum, frame):
    #    print("\nПолучен сигнал прерывания. Закрываем соединение...")
    #    self.reconnect()
    
    def log(self, end='\r'):
        try :
            #signal.signal(signal.SIGINT, self.signal_handler)
            m = DepolarizerMessage()
            mid = 0
            m.id = mid
            m.command = DepolarizerMessage.GET
            m.data_type = DepolarizerMessage.FMAP
            #Сделаем отдельный сокет для получения сообщений
            sock = Depolarizer.connect(self.host, self.port)
            #sock = self.sock
            Depolarizer.send(sock,m);
            m = Depolarizer.receive(sock);
            if m.status == DepolarizerMessage.OK:
                count = 0
                print("#{:^24}   {:^11}    {:^10}     {:6}       {:5}".format("time", "frequency", "energy", "speed", "atten"))
                while True:
                    #m = self.receive2()
                    m = Depolarizer.receive(sock);
                    if m!=None:
                        for st in m.states: 
                            t = datetime.datetime.fromtimestamp(st.timestamp*1e-9).isoformat(sep=' ', timespec='milliseconds')
                            n = st.harmonic_number
                            f = st.frequency
                            E = frequency2energy(f, n)
                            speed = st.speed
                            speed_kevs = frequency2energy(st.speed, 0)*1e3

                            print("{:24} {:11.3f} Hz {:10.3f} MeV {:6.3f} keV/s {:5} dB".format(t, f, E, speed_kevs, st.attenuation), end=end)
                            count += 1
                    else:
                        time.sleep(0.5)
        except:
            print("")
            #sock.shutdown(socket.SHUT_RDWR)
            sock.close()
            raise

if __name__ == '__main__':
    d = Depolarizer('vepp4-spin.inp.nsk.su',9090)
    print("power {}".format("ON" if d.is_on() else "OFF"))
    print("scan {}".format("ON" if d.is_scan() else "OFF"))
    print("attenuation ", d.get_attenuation())
    print("speed {:.3f} Hz".format(d.get_speed()))
    print(" step {:.3f} Hz".format(d.get_step()))
    print("initial frequency {:.3f} Hz".format(d.get_initial()))
    print("harmonic number: {}".format(d.get_harmonic_number()))

def scan(E, speed, attenuation, width=0.0, islog=True):
    # R = 2*m_e/(g-2) 
    R = 440.64845866025958 #MeV  // PDG-2021
    F0 = 818924.818
    s =  speed/R*F0
    w = width/R*F0
    harmonic = E//R
    initial = (E/R-harmonic)*F0
    d.set_harmonic_number(harmonic)
    d.set_attenuation(attenuation)
    d.set_speed(s)
    d.set_initial(initial)
    d.set_width(width)
    if d.start_scan():
        print("Scan from {:.3f} MeV {:}, speed {:.2f} keV/s, att. {} dB, harmonic {:d}".format(E,
              "up" if speed>0 else "down", 1000.*speed, int(attenuation), int(harmonic)))
    if  islog:
        try:
            d.log()
        except KeyboardInterrupt:
            print("\nStop scan")
            stop()


def scan_up(E,margin=0.8):
    scan(E-abs(margin), +0.001,28)

def scan_down(E,margin=0.8):
    scan(E+abs(margin), -0.001,28)

def stop():
    d.stop_scan()

def start():
    d.start_scan()

def on(f) :
    d.stop_scan()
    #d.set_attenuation(att)
    d.set_initial(f)
    d.set_speed(0.0)
    d.start_scan()

def utug(E, dE, speed, attenuation):
    print("Start utug with central {:.3f} MeV range +-{:.3f} MeV speed {:.2f} keV/s att. {} dB".format(E,dE, speed*1000, attenuation))
    #E - is real energy
    # dE is amplitude of sweeping
    T = 2*dE/speed
    try:
        while True:
            #scan up
            begin_time = time.time()
            scan(E-dE, +abs(speed), attenuation, islog=False)
            while time.time() < begin_time + T:
                time.sleep(1)
            d.stop_scan()
            time.sleep(1)
            #scan down
            begin_time = time.time()
            scan(E+dE, -abs(speed), attenuation,islog=False)
            while time.time() < begin_time + T:
                time.sleep(1)
            d.stop_scan()
            time.sleep(1)
    except KeyboardInterrupt:
        print("Stop utug")
        stop()

def shake(f, width, att, speed=0.0):
    d.set_attenuation(att)
    d.set_speed(speed)
    d.set_width(width)
    d.set_initial(f)
    d.start_scan()
    print("Start shaker with parameters: f={} Hz, width={} Hz, attenuation={} dB".format(d.get_frequency(), d.get_width() , d.get_attenuation()) )
    try:
        d.log()
    except KeyboardInterrupt:
        print("\nStop shaker")
        stop()

def test(t):
    while True:
        d.stop_scan();
        time.sleep(t);
        d.start_scan();
        time.sleep(t);
